@@ -1,4 +1,3 @@
-/* $Xorg: xprop.c,v 1.6 2001/02/09 02:05:56 xorgcvs Exp $ */
 /*
 
 Copyright 1990, 1998  The Open Group
@@ -27,7 +26,6 @@ other dealings in this Software without prior written authorization
 from The Open Group.
 
 */
-/* $XFree86: xc/programs/xprop/xprop.c,v 1.15 2003/09/24 02:43:38 dawes Exp $ */
 
 #include "config.h"
 
@@ -45,13 +43,15 @@ from The Open Group.
 #include <wctype.h>
 #endif
 #include <locale.h>
+#ifdef HAVE_LANGINFO_H
+#include <langinfo.h>
+#endif
 
 #ifndef HAVE_WCTYPE_H
 #define iswprint(x) isprint(x)
 #endif
 
 #include <X11/Xatom.h>
-#include <X11/Xmu/WinUtil.h>
 
 #include "dsimple.h"
 
@@ -86,7 +86,9 @@ Create_Thunk_List (void)
 {
     thunk *tptr;
 
-    tptr = (thunk *) Malloc(sizeof(thunk));
+    tptr = malloc(sizeof(thunk));
+    if (!tptr)
+	Fatal_Error("Out of memory!");
 
     tptr->thunk_count = 0;
 
@@ -108,7 +110,7 @@ Add_Thunk (thunk *list, thunk t)
 
     i = list->thunk_count;
 
-    list = (thunk *) realloc(list, (i+1)*sizeof(thunk));
+    list = realloc(list, (i+1)*sizeof(thunk));
     if (!list)
 	Fatal_Error("Out of memory!");
 
@@ -122,20 +124,6 @@ Add_Thunk (thunk *list, thunk t)
  * Misc. routines
  */
 
-static char *
-Copy_String (const char *string)
-{
-    char *new;
-    int length;
-
-    length = strlen(string) + 1;
-
-    new = (char *) Malloc(length);
-    memcpy(new, string, length);
-
-    return new;
-}
-
 static int
 Read_Char (FILE *stream)
 {
@@ -186,7 +174,10 @@ Read_Quoted (FILE *stream)
     }
     ptr++[0] = '\0';
 
-    return Copy_String(_large_buffer);
+    ptr = strdup(_large_buffer);
+    if (!ptr)
+	Fatal_Error("Out of memory!");
+    return ptr;
 }
 
 /*
@@ -296,7 +287,7 @@ Lookup_Formats (Atom atom, const char **format, const char **dformat)
 static void
 Add_Mapping (Atom atom, const char *format, const char *dformat)
 {
-    thunk t;
+    thunk t = {0};
 
     if (!_property_formats)
 	_property_formats = Create_Thunk_List();
@@ -356,7 +347,7 @@ typedef struct _propertyRec {
 "?m3(\t\twindow id # to use for icon: $4\n)"\
 "?m4(\t\tstarting position for icon: $5, $6\n)"\
 "?m6(\t\twindow id # of group leader: $8\n)"\
-"?m8(\t\tThe visible hint bit is set\n)"
+"?m8(\t\tThe urgency hint bit is set\n)"
 
 #define WM_ICON_SIZE_DFORMAT	":\n"\
 "\t\tminimum icon size: $0 by $1\n"\
@@ -407,6 +398,7 @@ static propertyRec windowPropTable[] = {
     {"RECTANGLE",	XA_RECTANGLE,	 "16iicc",    RECTANGLE_DFORMAT },
     {"RGB_COLOR_MAP",	XA_RGB_COLOR_MAP,"32xcccccccxx",RGB_COLOR_MAP_DFORMAT},
     {"STRING",		XA_STRING,	 "8s",	      0 },
+    {"UTF8_STRING",		0,	 "8u",	      0 },
     {"WINDOW",		XA_WINDOW,	 "32x",	      ": window id # $0+\n" },
     {"VISUALID",	XA_VISUALID,	 "32x",	      ": visual id # $0\n" },
     {"WM_COLORMAP_WINDOWS",	0,	 "32x",       ": window id # $0+\n"},
@@ -417,6 +409,7 @@ static propertyRec windowPropTable[] = {
     {"WM_NAME",		XA_WM_NAME,	 "8t",	      0 },
     {"WM_PROTOCOLS",		0,	 "32a",	      ": protocols  $0+\n"},
     {"WM_SIZE_HINTS",	XA_WM_SIZE_HINTS,"32mii",     WM_SIZE_HINTS_DFORMAT },
+    {"_NET_WM_ICON",            0,       "32o",        0 },
     {"WM_STATE",		0,	 "32cx",      WM_STATE_DFORMAT}
 };
 #undef ARC_DFORMAT
@@ -553,7 +546,9 @@ Read_Mappings (FILE *stream)
 	    Fatal_Error("Bad format file format.");
 
 	atom = Parse_Atom(name, False);
-	format = Copy_String(format_buffer);
+	format = strdup(format_buffer);
+	if (!format)
+	    Fatal_Error("Out of memory!");
 
 	Read_White_Space(stream);
 	dformat = D_DFORMAT;
@@ -680,7 +675,7 @@ _put_char (char c)
 }
 
 static void
-_format_char (char c)
+_format_char (char c, int unicode)
 {
     switch (c) {
       case '\\':
@@ -698,17 +693,21 @@ _format_char (char c)
 	break;
       default:
 	if (!c_isprint(c)) {
-	    _put_char('\\');
-	    snprintf(_buf_ptr, _buf_len, "%03o", (unsigned char) c);
-	    _buf_ptr += 3;
-	    _buf_len -= 3;
+	    if (unicode && (c & 0x80)) {
+		_put_char(c);
+	    } else {
+		_put_char('\\');
+		snprintf(_buf_ptr, _buf_len, "%03o", (unsigned char) c);
+		_buf_ptr += 3;
+		_buf_len -= 3;
+	    }
 	} else
 	  _put_char(c);
     }
 }
 
 static const char *
-Format_String (const char *string)
+Format_String (const char *string, int unicode)
 {
     char c;
 
@@ -717,7 +716,7 @@ Format_String (const char *string)
     _put_char('\"');
 
     while ((c = string++[0]))
-	_format_char(c);
+	_format_char(c, unicode);
 
     *_buf_ptr++ = '"';
     *_buf_ptr++ = '\0';
@@ -730,17 +729,128 @@ Format_Len_String (const char *string, int len)
     char *data;
     const char *result;
 
-    data = (char *) Malloc(len+1);
+    data = malloc(len+1);
+    if (!data)
+	Fatal_Error("Out of memory!");
 
     memcpy(data, string, len);
     data[len] = '\0';
 
-    result = Format_String(data);
+    result = Format_String(data, 0);
     free(data);
 
     return result;
 }  
 
+static int
+is_utf8_locale (void)
+{
+#ifdef HAVE_LANGINFO_H
+    char *charmap = nl_langinfo (CODESET);
+
+    return charmap && strcmp (charmap, "UTF-8") == 0;
+#else
+    return 0;
+#endif
+}
+
+static const char *
+Format_Icons (const unsigned long *icon, int len)
+{
+    char *result = NULL, *tail = NULL;
+    int alloced;
+    const unsigned long *end = icon + len / sizeof (unsigned long);
+
+    alloced = 0;
+
+    while (icon < end)
+    {
+	unsigned long width, height;
+	int w, h;
+	int offset;
+	
+	width = *icon++;
+	height = *icon++;
+
+	offset = (tail - result);
+	
+	alloced += 80;				/* For the header */
+	alloced += (width*4 + 8) * height;	/* For the rows (plus padding) */
+	
+	result = realloc (result, alloced);
+	if (!result)
+	    Fatal_Error("Out of memory!");
+	tail = &result[offset];
+
+	if (end - icon < width * height)
+	    break;
+
+	tail += sprintf (tail, "\tIcon (%lu x %lu):\n", width, height);
+
+	if (width > 144 || height > 144)
+	{
+	    tail += sprintf (tail, "\t(not shown)");
+	    icon += width * height;
+	    continue;
+	}
+	
+	for (h = 0; h < height; ++h)
+	{
+	    tail += sprintf (tail, "\t");
+	    
+	    for (w = 0; w < width; ++w)
+	    {
+		unsigned char a, r, g, b;
+		unsigned long pixel = *icon++;
+		unsigned long brightness;
+		
+		a = (pixel & 0xff000000) >> 24;
+		r = (pixel & 0x00ff0000) >> 16;
+		g = (pixel & 0x0000ff00) >> 8;
+		b = (pixel & 0x000000ff);
+		
+		brightness =
+		    (a / 255.0) * (1000 - ((299 * (r / 255.0)) +
+					   (587 * (g / 255.0)) +
+					   (114 * (b / 255.0))));
+
+		if (is_utf8_locale())
+		{
+		    static const char palette[][4] =
+		    {
+			" ",
+			"\342\226\221",		/* 25% */
+			"\342\226\222",		/* 50% */
+			"\342\226\223",		/* 75% */
+			"\342\226\210",		/* 100% */
+		    };
+		    int idx;
+
+		    idx = (brightness * ((sizeof (palette)/sizeof(palette[0])) - 1)) / 1000;
+
+		    tail += sprintf (tail, "%s", palette[idx]);
+		}
+		else
+		{
+		    static const char palette[] =
+			" .'`,^:\";~-_+<>i!lI?/\\|()1{}[]rcvunxzjftLCJUYXZO0Qoahkbdpqwm*WMB8&%$#@";
+		    int idx;
+		    
+		    idx = (brightness * (sizeof(palette) - 2)) / 1000;
+		    
+		    *tail++ = palette[idx];
+		}
+	    }
+
+	    tail += sprintf (tail, "\n");
+	}
+
+	tail += sprintf (tail, "\n");
+    }
+
+    return result;
+}
+
 static const char *
 Format_Len_Text (const char *string, int len, Atom encoding)
 {
@@ -794,6 +904,135 @@ Format_Len_Text (const char *string, int len, Atom encoding)
 	return Format_Len_String(string, len);
 }
 
+/*
+ * Validate a string as UTF-8 encoded according to RFC 3629
+ *
+ * Simply, a unicode code point (up to 21-bits long) is encoded as follows:
+ *
+ *    Char. number range  |        UTF-8 octet sequence
+ *       (hexadecimal)    |              (binary)
+ *    --------------------+---------------------------------------------
+ *    0000 0000-0000 007F | 0xxxxxxx
+ *    0000 0080-0000 07FF | 110xxxxx 10xxxxxx
+ *    0000 0800-0000 FFFF | 1110xxxx 10xxxxxx 10xxxxxx
+ *    0001 0000-0010 FFFF | 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx
+ *
+ * Validation is done left-to-right, and an error condition, if any, refers to
+ * only the left-most problem in the string.
+ *
+ * Return values:
+ *   UTF8_VALID: Valid UTF-8 encoded string
+ *   UTF8_OVERLONG: Using more bytes than needed for a code point
+ *   UTF8_SHORT_TAIL: Not enough bytes in a multi-byte sequence
+ *   UTF8_LONG_TAIL: Too many bytes in a multi-byte sequence
+ *   UTF8_FORBIDDEN_VALUE: Forbidden prefix or code point outside 0x10FFFF
+ */
+#define UTF8_VALID 0
+#define UTF8_FORBIDDEN_VALUE 1
+#define UTF8_OVERLONG 2
+#define UTF8_SHORT_TAIL 3
+#define UTF8_LONG_TAIL 4
+static int
+is_valid_utf8 (const char *string, int len)
+{
+    unsigned long codepoint = 0;
+    int rem, i;
+    unsigned char c;
+
+    rem = 0;
+    for (i = 0; i < len; i++) {
+	c = (unsigned char) string[i];
+
+	/* Order of type check:
+	 *   - Single byte code point
+	 *   - Non-starting byte of multi-byte sequence
+	 *   - Start of 2-byte sequence
+	 *   - Start of 3-byte sequence
+	 *   - Start of 4-byte sequence
+	 */
+	if (!(c & 0x80)) {
+	    if (rem > 0) return UTF8_SHORT_TAIL;
+	    rem = 0;
+	    codepoint = c;
+	} else if ((c & 0xC0) == 0x80) {
+	    if (rem == 0) return UTF8_LONG_TAIL;
+	    rem--;
+	    codepoint |= (c & 0x3F) << (rem * 6);
+	    if (codepoint == 0) return UTF8_OVERLONG;
+	} else if ((c & 0xE0) == 0xC0) {
+	    if (rem > 0) return UTF8_SHORT_TAIL;
+	    rem = 1;
+	    codepoint = (c & 0x1F) << 6;
+	    if (codepoint == 0) return UTF8_OVERLONG;
+	} else if ((c & 0xF0) == 0xE0) {
+	    if (rem > 0) return UTF8_SHORT_TAIL;
+	    rem = 2;
+	    codepoint = (c & 0x0F) << 12;
+	} else if ((c & 0xF8) == 0xF0) {
+	    if (rem > 0) return UTF8_SHORT_TAIL;
+	    rem = 3;
+	    codepoint = (c & 0x07) << 18;
+	    if (codepoint > 0x10FFFF) return UTF8_FORBIDDEN_VALUE;
+	} else
+	    return UTF8_FORBIDDEN_VALUE;
+    }
+
+    return UTF8_VALID;
+}
+
+static const char *
+Format_Len_Unicode (const char *string, int len)
+{
+    char *data;
+    const char *result, *error;
+    int len2;
+
+    int validity = is_valid_utf8(string, len);
+
+    if (validity != UTF8_VALID) {
+	switch (validity) {
+	  case UTF8_FORBIDDEN_VALUE:
+	    error = "<Invalid UTF-8 string: Forbidden value> "; break;
+	  case UTF8_OVERLONG:
+	    error = "<Invalid UTF-8 string: Overlong encoding> "; break;
+	  case UTF8_SHORT_TAIL:
+	    error = "<Invalid UTF-8 string: Tail too short> "; break;
+	  case UTF8_LONG_TAIL:
+	    error = "<Invalid UTF-8 string: Tail too long> "; break;
+	  default:
+	    error = "<Invalid UTF-8 string: Unknown error>"; break;
+	}
+
+	result = Format_Len_String(string, len);
+	len2 = strlen(result);
+	data = malloc(len2+1);
+	if (!data)
+	    Fatal_Error("Out of memory!");
+	memcpy(data, result, len2+1);
+
+	memcpy(_formatting_buffer, error, strlen(error)+1);
+	strcat(_formatting_buffer, data);
+	free(data);
+
+	return _formatting_buffer;
+    }
+
+    if (!is_utf8_locale())
+	return Format_Len_String(string, len);
+
+    data = malloc(len+1);
+    if (!data)
+	Fatal_Error("Out of memory!");
+
+    memcpy(data, string, len);
+    data[len] = '\0';
+
+    result = Format_String(data, 1);
+    free(data);
+
+    return result;
+}
+
 /*
  *
  * The Format Manager: a group of routines to manage "formats"
@@ -846,6 +1085,8 @@ Format_Thunk (thunk t, char format_char)
     switch (format_char) {
       case 's':
 	return Format_Len_String(t.extra_value, (int)t.value);
+      case 'u':
+	return Format_Len_Unicode(t.extra_value, (int)t.value);
       case 't':
 	return Format_Len_Text(t.extra_value, (int)t.value, t.extra_encoding);
       case 'x':
@@ -860,6 +1101,8 @@ Format_Thunk (thunk t, char format_char)
 	return Format_Mask_Word(value);
       case 'a':
 	return Format_Atom(value);
+      case 'o':
+	  return Format_Icons((const unsigned long *)t.extra_value, (int)t.value);
       default:
 	Fatal_Error("bad format character: %c", format_char);
     }
@@ -1113,11 +1356,25 @@ Extract_Len_String (const char **pointer, int *length, int size, const char **st
     return len;
 }
 
+static long
+Extract_Icon (const char **pointer, int *length, int size, const char **icon)
+{
+    int len = 0;
+
+    if (size != 32)
+	Fatal_Error("can't use format character 'o' with any size except 32.");
+
+    len = *length;
+    *icon = *pointer;
+    *length = 0;
+    return len;
+}
+
 static thunk *
 Break_Down_Property (const char *pointer, int length, Atom type, const char *format, int size)
 {
     thunk *thunks;
-    thunk t;
+    thunk t = {0};
     int i;
     char format_char;
 
@@ -1126,12 +1383,15 @@ Break_Down_Property (const char *pointer, int length, Atom type, const char *for
 
     while (length >= size/8) {
 	format_char = Get_Format_Char(format, i);
-	if (format_char == 's')
+	if (format_char == 's' || format_char == 'u')
 	    t.value = Extract_Len_String(&pointer,&length,size,&t.extra_value);
 	else if (format_char == 't') {
 	    t.extra_encoding = type;
 	    t.value = Extract_Len_String(&pointer,&length,size,&t.extra_value);
-	} else
+	}
+	else if (format_char == 'o')
+	    t.value = Extract_Icon (&pointer,&length,size,&t.extra_value);
+	else
 	    t.value = Extract_Value(&pointer,&length,size,format_char=='i');
 	thunks = Add_Thunk(thunks, t);
 	i++;
@@ -1288,7 +1548,7 @@ static thunk *
 Handle_Prop_Requests (int argc, char **argv)
 {
     char *format, *dformat, *prop;
-    thunk *thunks, t;
+    thunk *thunks, t = {0};
 
     /* if no prop referenced, by default list all properties for given window */
     if (!argc) {
@@ -1597,6 +1857,22 @@ Parse_Format_Mapping (int *argc, char ***argv)
 
 static int spy = 0;
 
+static int (*old_error_handler)(Display *dpy, XErrorEvent *ev);
+
+static int spy_error_handler(Display *dpy, XErrorEvent *ev)
+{
+    if (ev->error_code == BadWindow || ev->error_code == BadMatch) {
+	/* Window was destroyed */
+	puts("");
+	exit(0);
+    }
+
+    if (old_error_handler)
+	return old_error_handler(dpy, ev);
+
+    return 0;
+}
+
 int
 main (int argc, char **argv)
 {
@@ -1671,7 +1947,7 @@ main (int argc, char **argv)
 	    continue;
 	}
 	if (!strcmp(argv[0], "-remove")) {
-	    thunk t;
+	    thunk t = {0};
 	    if (++argv, --argc == 0) usage();
 	    t.propname = argv[0];
 	    if (remove_props == NULL) remove_props = Create_Thunk_List();
@@ -1679,7 +1955,7 @@ main (int argc, char **argv)
 	    continue;
 	}
 	if (!strcmp(argv[0], "-set")) {
-	    thunk t;
+	    thunk t = {0};
 	    if (argc < 3) usage();
 	    t.propname = argv[1];
 	    t.extra_value = argv[2];
@@ -1702,19 +1978,8 @@ main (int argc, char **argv)
     if ((remove_props != NULL || set_props != NULL) && argc > 0)
 	usage();
 
-    if (target_win == None) {
-	target_win = Select_Window(dpy);
-	if (target_win != None && !frame_only) {
-	    Window root;
-	    int dummyi;
-	    unsigned int dummy;
-
-	    if (XGetGeometry (dpy, target_win, &root, &dummyi, &dummyi,
-			      &dummy, &dummy, &dummy, &dummy)
-		&& target_win != root)
-		target_win = XmuClientWindow (dpy, target_win);
-	}
-    }
+    if (target_win == None)
+	target_win = Select_Window(dpy, !frame_only);
 
     if (remove_props != NULL) {
 	int count;
@@ -1750,9 +2015,15 @@ main (int argc, char **argv)
 	XEvent event;
 	const char *format, *dformat;
 	
-	XSelectInput(dpy, target_win, PropertyChangeMask);
+	XSelectInput(dpy, target_win, PropertyChangeMask | StructureNotifyMask);
+ 	old_error_handler = XSetErrorHandler(spy_error_handler);
 	for (;;) {
+	    fflush(stdout);
 	    XNextEvent(dpy, &event);
+ 	    if (event.type == DestroyNotify)
+ 		break;
+ 	    if (event.type != PropertyNotify)
+ 		continue;
 	    format = dformat = NULL;
 	    if (props) {
 		int i;
